# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron.tests import base
from neutron_lib.api.definitions import bgp as bgp_ext
from neutron_lib.plugins import directory

from neutron_dynamic_routing.api.rpc.handlers import bgp_speaker_rpc


class TestBgpSpeakerRpcCallback(base.BaseTestCase):

    def setUp(self):
        super(TestBgpSpeakerRpcCallback, self).setUp()
        self.plugin = mock.Mock()
        directory.add_plugin(bgp_ext.ALIAS, self.plugin)
        self.callback = bgp_speaker_rpc.BgpSpeakerRpcCallback()

    def test_get_bgp_speaker_info(self):
        self.callback.get_bgp_speaker_info(mock.Mock(),
                                           bgp_speaker_id='id1')
        self.assertIsNotNone(len(self.plugin.mock_calls))

    def test_get_bgp_peer_info(self):
        self.callback.get_bgp_peer_info(mock.Mock(),
                                        bgp_peer_id='id1')
        self.assertIsNotNone(len(self.plugin.mock_calls))

    def test_get_bgp_speakers(self):
        self.callback.get_bgp_speakers(mock.Mock(),
                                       host='host')
        self.assertIsNotNone(len(self.plugin.mock_calls))



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import constants as lib_consts

from neutron.tests import base

from neutron_dynamic_routing.services.bgp.agent.driver import exceptions as bgp_driver_exc  # noqa
from neutron_dynamic_routing.services.bgp.agent.driver import utils as bgp_driver_utils  # noqa
from neutron_dynamic_routing.services.bgp.common import constants as bgp_consts  # noqa

FAKE_IP = '2.2.2.5'
FAKE_IPV6 = '2001:db8::'
FAKE_LOCAL_AS = 12345
FAKE_OS_KEN_SPEAKER = {}
EXC_INV_PARAMTYPE = "Parameter %(param)s must be of %(param_type)s type."
EXC_INV_PARAMRANGE = "%(param)s must be in %(range)s range."
EXC_PASSWORD_NOTSPEC = "Password not specified for authentication " + \
                       "type=%(auth_type)s."
EXC_INV_AUTHTYPE = "Authentication type not supported. Requested " + \
                   "type=%(auth_type)s."


class TestValidateMethod(base.BaseTestCase):

    def setUp(self):
        super(TestValidateMethod, self).setUp()

    def test_validate_as_num_with_valid_as_num(self):
        self.assertIsNone(bgp_driver_utils.validate_as_num('local_as',
                                                           64512))

    def test_validate_as_num_with_string_as_num(self):
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamType,
                EXC_INV_PARAMTYPE % {'param': 'local_as',
                                     'param_type': 'integer'}):
            bgp_driver_utils.validate_as_num('local_as', '64512')

    def test_validate_as_num_with_invalid_max_range(self):
        allowed_range = ('\\[' +
                         str(bgp_consts.MIN_ASNUM) + '-' +
                         str(bgp_consts.MAX_4BYTE_ASNUM) +
                         '\\]')
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamRange,
                EXC_INV_PARAMRANGE % {'param': 'local_as',
                                      'range': allowed_range}):
            bgp_driver_utils.validate_as_num('local_as',
                                             bgp_consts.MAX_4BYTE_ASNUM + 1)

    def test_validate_as_num_with_invalid_min_range(self):
        allowed_range = ('\\[' +
                         str(bgp_consts.MIN_ASNUM) + '-' +
                         str(bgp_consts.MAX_4BYTE_ASNUM) +
                         '\\]')
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamRange,
                EXC_INV_PARAMRANGE % {'param': 'local_as',
                                      'range': allowed_range}):
            bgp_driver_utils.validate_as_num('local_as', 0)

    def test_validate_auth_with_valid_auth_type(self):
        passwords = [None, 'password']
        for (auth_type, passwd) in zip(bgp_consts.SUPPORTED_AUTH_TYPES,
                                       passwords):
            self.assertIsNone(bgp_driver_utils.validate_auth(auth_type,
                                                             passwd))

    def test_validate_auth_with_integer_password(self):
        auth_type = bgp_consts.SUPPORTED_AUTH_TYPES[1]
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamType,
                EXC_INV_PARAMTYPE % {'param': '12345',
                                     'param_type': 'string'}):
            bgp_driver_utils.validate_auth(auth_type, 12345)

    def test_validate_auth_with_invalid_auth_type(self):
        auth_type = 'abcde'
        with self.assertRaisesRegex(
                bgp_driver_exc.InvaildAuthType,
                EXC_INV_AUTHTYPE % {'auth_type': auth_type}):
            bgp_driver_utils.validate_auth(auth_type, 'password')

    def test_validate_auth_with_not_none_auth_type_and_none_password(self):
        auth_type = bgp_consts.SUPPORTED_AUTH_TYPES[1]
        with self.assertRaisesRegex(
                bgp_driver_exc.PasswordNotSpecified,
                EXC_PASSWORD_NOTSPEC % {'auth_type': auth_type}):
            bgp_driver_utils.validate_auth(auth_type, None)

    def test_validate_auth_with_none_auth_type_and_not_none_password(self):
        auth_type = None
        with self.assertRaisesRegex(
                bgp_driver_exc.InvaildAuthType,
                EXC_INV_AUTHTYPE % {'auth_type': auth_type}):
            bgp_driver_utils.validate_auth(auth_type, 'password')

    def test_validate_ip_addr_with_ipv4_address(self):
        self.assertEqual(lib_consts.IP_VERSION_4,
                         bgp_driver_utils.validate_ip_addr(FAKE_IP))

    def test_validate_ip_addr_with_ipv6_address(self):
        self.assertEqual(lib_consts.IP_VERSION_6,
                         bgp_driver_utils.validate_ip_addr(FAKE_IPV6))

    def test_validate_ip_addr_with_integer_ip(self):
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamType,
                EXC_INV_PARAMTYPE % {'param': '12345',
                                     'param_type': 'ip-address'}):
            bgp_driver_utils.validate_ip_addr(12345)

    def test_validate_ip_addr_with_invalid_ipv4_type(self):
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamType,
                EXC_INV_PARAMTYPE % {'param': '1.2.3.a',
                                     'param_type': 'ip-address'}):
            bgp_driver_utils.validate_ip_addr('1.2.3.a')

    def test_validate_ip_addr_with_invalid_ipv6_type(self):
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamType,
                EXC_INV_PARAMTYPE % {'param': '2001:db8::ggg',
                                     'param_type': 'ip-address'}):
            bgp_driver_utils.validate_ip_addr('2001:db8::ggg')

    def test_validate_string_with_string(self):
        self.assertIsNone(bgp_driver_utils.validate_string(FAKE_IP))

    def test_validate_string_with_integer_param(self):
        with self.assertRaisesRegex(
                bgp_driver_exc.InvalidParamType,
                EXC_INV_PARAMTYPE % {'param': '12345',
                                     'param_type': 'string'}):
            bgp_driver_utils.validate_string(12345)


class TestBgpMultiSpeakerCache(base.BaseTestCase):

    def setUp(self):
        super(TestBgpMultiSpeakerCache, self).setUp()
        self.expected_cache = {FAKE_LOCAL_AS: FAKE_OS_KEN_SPEAKER}
        self.bs_cache = bgp_driver_utils.BgpMultiSpeakerCache()

    def test_put_bgp_speaker(self):
        self.bs_cache.put_bgp_speaker(FAKE_LOCAL_AS, FAKE_OS_KEN_SPEAKER)
        self.assertEqual(self.expected_cache, self.bs_cache.cache)

    def test_remove_bgp_speaker(self):
        self.bs_cache.put_bgp_speaker(FAKE_LOCAL_AS, FAKE_OS_KEN_SPEAKER)
        self.assertEqual(1, len(self.bs_cache.cache))
        self.bs_cache.remove_bgp_speaker(FAKE_LOCAL_AS)
        self.assertEqual(0, len(self.bs_cache.cache))

    def test_get_bgp_speaker(self):
        self.bs_cache.put_bgp_speaker(FAKE_LOCAL_AS, FAKE_OS_KEN_SPEAKER)
        self.assertEqual(
            FAKE_OS_KEN_SPEAKER,
            self.bs_cache.get_bgp_speaker(FAKE_LOCAL_AS))

    def test_get_hosted_bgp_speakers_count(self):
        self.bs_cache.put_bgp_speaker(FAKE_LOCAL_AS, FAKE_OS_KEN_SPEAKER)
        self.assertEqual(1, self.bs_cache.get_hosted_bgp_speakers_count())



# Copyright 2016 Hewlett Packard Development Coompany LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from neutron_lib.api.definitions import bgp as bgp_api_def
from neutron_lib.api import extensions as api_ext
from neutron_lib import exceptions as n_exc

from neutron.api.v2 import resource_helper as rh

from neutron_dynamic_routing._i18n import _


# Dynamic Routing Exceptions
class BgpSpeakerNotFound(n_exc.NotFound):
    message = _("BGP speaker %(id)s could not be found.")


class BgpPeerNotFound(n_exc.NotFound):
    message = _("BGP peer %(id)s could not be found.")


class BgpPeerNotAuthenticated(n_exc.NotFound):
    message = _("BGP peer %(bgp_peer_id)s not authenticated.")


class BgpSpeakerPeerNotAssociated(n_exc.NotFound):
    message = _("BGP peer %(bgp_peer_id)s is not associated with "
                "BGP speaker %(bgp_speaker_id)s.")


class BgpSpeakerNetworkNotAssociated(n_exc.NotFound):
    message = _("Network %(network_id)s is not associated with "
                "BGP speaker %(bgp_speaker_id)s.")


class BgpSpeakerNetworkBindingError(n_exc.Conflict):
    message = _("Network %(network_id)s is already bound to BgpSpeaker "
                "%(bgp_speaker_id)s.")


class NetworkNotBound(n_exc.NotFound):
    message = _("Network %(network_id)s is not bound to a BgpSpeaker.")


class DuplicateBgpPeerIpException(n_exc.Conflict):
    message = _("BGP Speaker %(bgp_speaker_id)s is already configured to "
                "peer with a BGP Peer at %(peer_ip)s, it cannot peer with "
                "BGP Peer %(bgp_peer_id)s.")


class InvalidBgpPeerMd5Authentication(n_exc.BadRequest):
    message = _("A password must be supplied when using auth_type md5.")


class NetworkNotBoundForIpVersion(NetworkNotBound):
    message = _("Network %(network_id)s is not bound to a IPv%(ip_version)s "
                "BgpSpeaker.")


class Bgp(api_ext.APIExtensionDescriptor):
    api_definition = bgp_api_def

    @classmethod
    def get_resources(cls):
        plural_mappings = rh.build_plural_mappings(
            {}, bgp_api_def.RESOURCE_ATTRIBUTE_MAP)
        exts = rh.build_resource_info(plural_mappings,
                                      bgp_api_def.RESOURCE_ATTRIBUTE_MAP,
                                      bgp_api_def.ALIAS,
                                      action_map=bgp_api_def.ACTION_MAP)

        return exts



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""change size of as number

Revision ID: a589fdb5724c
Revises: 4cf8bc3edb66
Create Date: 2017-08-31 13:50:28.324422

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = 'a589fdb5724c'
down_revision = '4cf8bc3edb66'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.QUEENS]


def upgrade():
    op.alter_column('bgp_speakers', 'local_as', nullable=False,
                    type_=sa.BigInteger())
    op.alter_column('bgp_peers', 'remote_as', nullable=False,
                    type_=sa.BigInteger())



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from neutron.api import extensions
from neutron.api.v2 import resource
from neutron import wsgi
from neutron_lib.api.definitions import bgp as bgp_ext
from neutron_lib.api.definitions import bgp_dragentscheduler as api_def
from neutron_lib.api import extensions as api_extensions
from neutron_lib.api import faults
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import agent as agent_exc
from neutron_lib.plugins import directory
from oslo_log import log as logging
import webob

from neutron_dynamic_routing._i18n import _, _LE


LOG = logging.getLogger(__name__)


class DrAgentInvalid(agent_exc.AgentNotFound):
    message = _("BgpDrAgent %(id)s is invalid or has been disabled.")


class DrAgentNotHostingBgpSpeaker(n_exc.NotFound):
    message = _("BGP speaker %(bgp_speaker_id)s is not hosted "
                "by the BgpDrAgent %(agent_id)s.")


class DrAgentAssociationError(n_exc.Conflict):
    message = _("BgpDrAgent %(agent_id)s is already associated "
                "to a BGP speaker.")


class BgpSpeakerRescheduleError(n_exc.Conflict):
    message = _("Failed rescheduling %(bgp_speaker_id)s: "
                "%(failure_reason)s.")


class BgpDrSchedulerController(wsgi.Controller):
    """Schedule BgpSpeaker for a BgpDrAgent"""
    def get_plugin(self):
        plugin = directory.get_plugin(bgp_ext.ALIAS)
        if not plugin:
            LOG.error(_LE('No plugin for BGP routing registered'))
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)
        return plugin

    def index(self, request, **kwargs):
        plugin = self.get_plugin()
        return plugin.list_bgp_speaker_on_dragent(
            request.context, kwargs['agent_id'])

    def create(self, request, body, **kwargs):
        plugin = self.get_plugin()
        return plugin.add_bgp_speaker_to_dragent(
            request.context,
            kwargs['agent_id'],
            body['bgp_speaker_id'])

    def delete(self, request, id, **kwargs):
        plugin = self.get_plugin()
        return plugin.remove_bgp_speaker_from_dragent(
            request.context, kwargs['agent_id'], id)


class BgpDrAgentController(wsgi.Controller):
    def get_plugin(self):
        plugin = directory.get_plugin(bgp_ext.ALIAS)
        if not plugin:
            LOG.error(_LE('No plugin for BGP routing registered'))
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)
        return plugin

    def index(self, request, **kwargs):
        plugin = directory.get_plugin(bgp_ext.ALIAS)
        return plugin.list_dragent_hosting_bgp_speaker(
            request.context, kwargs['bgp_speaker_id'])


class Bgp_dragentscheduler(api_extensions.APIExtensionDescriptor):
    """Extension class supporting Dynamic Routing scheduler.
    """

    api_definition = api_def

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        exts = []
        parent = dict(member_name="agent",
                      collection_name="agents")

        controller = resource.Resource(BgpDrSchedulerController(),
                                       faults.FAULT_MAP)
        exts.append(extensions.ResourceExtension(api_def.BGP_DRINSTANCES,
                                                 controller, parent))

        parent = dict(member_name="bgp_speaker",
                      collection_name="bgp-speakers")
        controller = resource.Resource(BgpDrAgentController(),
                                       faults.FAULT_MAP)
        exts.append(extensions.ResourceExtension(api_def.BGP_DRAGENTS,
                                                 controller, parent))
        return exts


class BgpDrSchedulerPluginBase(object, metaclass=abc.ABCMeta):
    """REST API to operate BGP dynamic routing agent scheduler.

    All the methods must be executed in admin context.
    """
    def get_plugin_description(self):
        return "Neutron BGP dynamic routing scheduler Plugin"

    def get_plugin_type(self):
        return bgp_ext.ALIAS

    @abc.abstractmethod
    def add_bgp_speaker_to_dragent(self, context, agent_id, speaker_id):
        pass

    @abc.abstractmethod
    def remove_bgp_speaker_from_dragent(self, context, agent_id, speaker_id):
        pass

    @abc.abstractmethod
    def list_dragent_hosting_bgp_speaker(self, context, speaker_id):
        pass

    @abc.abstractmethod
    def list_bgp_speaker_on_dragent(self, context, agent_id):
        pass

    @abc.abstractmethod
    def get_bgp_speakers_for_agent_host(self, context, host):
        pass

    @abc.abstractmethod
    def get_bgp_speaker_by_speaker_id(self, context, speaker_id):
        pass

    @abc.abstractmethod
    def get_bgp_peer_by_peer_id(self, context, bgp_peer_id):
        pass



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from neutron_dynamic_routing._i18n import _

BGP_DRIVER_OPTS = [
    cfg.StrOpt('bgp_speaker_driver',
               help=_("BGP speaker driver class to be instantiated."))
]

BGP_PROTO_CONFIG_OPTS = [
    cfg.StrOpt('bgp_router_id',
               help=_("32-bit BGP identifier, typically an IPv4 address "
                      "owned by the system running the BGP DrAgent."))
]



# Copyright (c) 2016 Hewlett Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api.definitions import bgp
from neutron_lib import context
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_utils import importutils

from neutron.extensions import agent
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_db_base_plugin
from neutron.tests.unit.extensions import test_agent

from neutron_dynamic_routing.db import bgp_db
from neutron_dynamic_routing.db import bgp_dragentscheduler_db as bgp_dras_db
from neutron_dynamic_routing.extensions import bgp_dragentscheduler as bgp_dras_ext  # noqa
from neutron_dynamic_routing.tests.common import helpers
from neutron_dynamic_routing.tests.unit.db import test_bgp_db

from webob import exc


class BgpDrSchedulerTestExtensionManager(object):

    def get_resources(self):
        return (agent.Agent.get_resources() +
                bgp_dras_ext.Bgp_dragentscheduler.get_resources())

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class TestBgpDrSchedulerPlugin(bgp_db.BgpDbMixin,
                               bgp_dras_db.BgpDrAgentSchedulerDbMixin):

    bgp_drscheduler = importutils.import_object(
        cfg.CONF.bgp_drscheduler_driver)

    supported_extension_aliases = ["bgp_dragent_scheduler"]

    def get_plugin_description(self):
        return ("BGP dynamic routing service Plugin test class that test "
                "BGP speaker functionality, with scheduler.")


class BgpDrSchedulingTestCase(test_agent.AgentDBTestMixIn,
                              test_bgp_db.BgpEntityCreationMixin):

    def test_schedule_bgp_speaker(self):
        """Test happy path over full scheduling cycle."""
        with self.bgp_speaker(4, 1234) as ri:
            bgp_speaker_id = ri['id']
            helpers.register_bgp_dragent(host='host1')
            agent = self._list('agents')['agents'][0]
            agent_id = agent['id']

            data = {'bgp_speaker_id': bgp_speaker_id}
            req = self.new_create_request('agents', data, self.fmt,
                                          agent_id, 'bgp-drinstances')
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPCreated.code, res.status_int)

            req_show = self.new_show_request('agents', agent_id, self.fmt,
                                             'bgp-drinstances')
            res = req_show.get_response(self.ext_api)
            self.assertEqual(exc.HTTPOk.code, res.status_int)
            res = self.deserialize(self.fmt, res)
            self.assertIn('bgp_speakers', res)
            self.assertTrue(bgp_speaker_id,
                            res['bgp_speakers'][0]['id'])

            req = self.new_delete_request('agents',
                                          agent_id,
                                          self.fmt,
                                          'bgp-drinstances',
                                          bgp_speaker_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPNoContent.code, res.status_int)

            res = req_show.get_response(self.ext_api)
            self.assertEqual(exc.HTTPOk.code, res.status_int)
            res = self.deserialize(self.fmt, res)
            self.assertIn('bgp_speakers', res)
            self.assertEqual([], res['bgp_speakers'])

    def test_schedule_bgp_speaker_on_invalid_agent(self):
        """Test error while scheduling BGP speaker on an invalid agent."""
        with self.bgp_speaker(4, 1234) as ri:
            bgp_speaker_id = ri['id']
            self._register_l3_agent(host='host1')  # Register wrong agent
            agent = self._list('agents')['agents'][0]
            data = {'bgp_speaker_id': bgp_speaker_id}
            req = self.new_create_request(
                'agents', data, self.fmt,
                agent['id'], 'bgp-drinstances')
            res = req.get_response(self.ext_api)

            # Raises an AgentNotFound exception if the agent is invalid
            self.assertEqual(exc.HTTPNotFound.code, res.status_int)

    def test_schedule_bgp_speaker_twice_on_same_agent(self):
        """Test error if a BGP speaker is scheduled twice on same agent"""
        with self.bgp_speaker(4, 1234) as ri:
            bgp_speaker_id = ri['id']
            helpers.register_bgp_dragent(host='host1')
            agent = self._list('agents')['agents'][0]
            data = {'bgp_speaker_id': bgp_speaker_id}
            req = self.new_create_request(
                'agents', data, self.fmt,
                agent['id'], 'bgp-drinstances')
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPCreated.code, res.status_int)

            # Try second time, should raise conflict
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPConflict.code, res.status_int)

    def test_schedule_bgp_speaker_on_two_different_agents(self):
        """Test that a BGP speaker can be associated to two agents."""
        with self.bgp_speaker(4, 1234) as ri:
            bgp_speaker_id = ri['id']
            helpers.register_bgp_dragent(host='host1')
            helpers.register_bgp_dragent(host='host2')
            data = {'bgp_speaker_id': bgp_speaker_id}

            agent1 = self._list('agents')['agents'][0]
            req = self.new_create_request(
                'agents', data, self.fmt,
                agent1['id'], 'bgp-drinstances')
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPCreated.code, res.status_int)

            agent2 = self._list('agents')['agents'][1]
            req = self.new_create_request(
                'agents', data, self.fmt,
                agent2['id'], 'bgp-drinstances')
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPCreated.code, res.status_int)

    def test_schedule_multi_bgp_speaker_on_one_dragent(self):
        """Test only one BGP speaker can be associated to one dragent."""
        with self.bgp_speaker(4, 1) as ri1, self.bgp_speaker(4, 2) as ri2:
            helpers.register_bgp_dragent(host='host1')

            agent = self._list('agents')['agents'][0]
            data = {'bgp_speaker_id': ri1['id']}
            req = self.new_create_request(
                'agents', data, self.fmt,
                agent['id'], 'bgp-drinstances')
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPCreated.code, res.status_int)

            data = {'bgp_speaker_id': ri2['id']}
            req = self.new_create_request(
                'agents', data, self.fmt,
                agent['id'], 'bgp-drinstances')
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPConflict.code, res.status_int)

    def test_non_scheduled_bgp_speaker_binding_removal(self):
        """Test exception while removing an invalid binding."""
        with self.bgp_speaker(4, 1234) as ri1:
            helpers.register_bgp_dragent(host='host1')
            agent = self._list('agents')['agents'][0]
            agent_id = agent['id']
            self.assertRaises(bgp_dras_ext.DrAgentNotHostingBgpSpeaker,
                              self.bgp_plugin.remove_bgp_speaker_from_dragent,
                              self.context, agent_id, ri1['id'])


class BgpDrPluginSchedulerTests(test_db_base_plugin.NeutronDbPluginV2TestCase,
                                BgpDrSchedulingTestCase):

    def setUp(self, plugin=None, ext_mgr=None, service_plugins=None):
        if not plugin:
            plugin = ('neutron_dynamic_routing.tests.unit.db.'
                      'test_bgp_dragentscheduler_db.TestBgpDrSchedulerPlugin')
        if not service_plugins:
            service_plugins = {bgp.ALIAS:
                               'neutron_dynamic_routing.services.bgp.'
                               'bgp_plugin.BgpPlugin'}

        ext_mgr = ext_mgr or BgpDrSchedulerTestExtensionManager()
        super(BgpDrPluginSchedulerTests, self).setUp(
            plugin=plugin, ext_mgr=ext_mgr, service_plugins=service_plugins)
        self.bgp_plugin = directory.get_plugin(bgp.ALIAS)
        self.context = context.get_admin_context()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet


eventlet.monkey_patch()
# Monkey patch the original current_thread to use the up-to-date _active
# global variable. See https://bugs.launchpad.net/bugs/1863021 and
# https://github.com/eventlet/eventlet/issues/592
import __original_module_threading as orig_threading  # noqa
import threading  # noqa
orig_threading.current_thread.__globals__['_active'] = threading._active



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

import neutron_dynamic_routing.services.bgp.agent.config


def list_bgp_agent_opts():
    return [
        ('bgp',
         itertools.chain(
             neutron_dynamic_routing.services.bgp.agent.
             config.BGP_DRIVER_OPTS,
             neutron_dynamic_routing.services.bgp.agent.
             config.BGP_PROTO_CONFIG_OPTS)
         )
    ]



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_ken.services.protocols.bgp import bgpspeaker
from os_ken.services.protocols.bgp.rtconf.neighbors import CONNECT_MODE_ACTIVE
from oslo_log import log as logging
from oslo_utils import encodeutils

from neutron_dynamic_routing._i18n import _LE, _LI
from neutron_dynamic_routing.services.bgp.agent.driver import base
from neutron_dynamic_routing.services.bgp.agent.driver import exceptions as bgp_driver_exc  # noqa
from neutron_dynamic_routing.services.bgp.agent.driver import utils

LOG = logging.getLogger(__name__)


# Function for logging BGP peer and path changes.
def bgp_peer_down_cb(remote_ip, remote_as):
    LOG.info(_LI('BGP Peer %(peer_ip)s for remote_as=%(peer_as)d went DOWN.'),
             {'peer_ip': remote_ip, 'peer_as': remote_as})


def bgp_peer_up_cb(remote_ip, remote_as):
    LOG.info(_LI('BGP Peer %(peer_ip)s for remote_as=%(peer_as)d is UP.'),
             {'peer_ip': remote_ip, 'peer_as': remote_as})


def best_path_change_cb(event):
    LOG.info(_LI("Best path change observed. cidr=%(prefix)s, "
                 "nexthop=%(nexthop)s, remote_as=%(remote_as)d, "
                 "is_withdraw=%(is_withdraw)s"),
             {'prefix': event.prefix, 'nexthop': event.nexthop,
              'remote_as': event.remote_as,
              'is_withdraw': event.is_withdraw})


class OsKenBgpDriver(base.BgpDriverBase):
    """BGP speaker implementation via os-ken."""

    def __init__(self, cfg):
        LOG.info(_LI('Initializing os-ken driver for BGP functionality.'))
        self._read_config(cfg)

        # Note: Even though os-ken can only support one BGP speaker as of now,
        # we have tried making the framework generic for the future purposes.
        self.cache = utils.BgpMultiSpeakerCache()

    def _read_config(self, cfg):
        if cfg is None or cfg.bgp_router_id is None:
            # If either cfg or router_id is not specified, raise voice
            LOG.error(_LE('BGP router-id MUST be specified for the correct '
                          'functional working.'))
        else:
            self.routerid = cfg.bgp_router_id
            LOG.info(_LI('Initialized os-ken BGP Speaker driver interface '
                         'with bgp_router_id=%s'), self.routerid)

    def add_bgp_speaker(self, speaker_as):
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if curr_speaker is not None:
            raise bgp_driver_exc.BgpSpeakerAlreadyScheduled(
                                                    current_as=speaker_as,
                                                    rtid=self.routerid)

        # os-ken can only support One speaker
        if self.cache.get_hosted_bgp_speakers_count() == 1:
            raise bgp_driver_exc.BgpSpeakerMaxScheduled(count=1)

        # Validate input parameters.
        # speaker_as must be an integer in the allowed range.
        utils.validate_as_num('local_as', speaker_as)

        # Notify os-ken about BGP Speaker addition.
        # Please note: Since, only the route-advertisement support is
        # implemented we are explicitly setting the bgp_server_port
        # attribute to 0 which disables listening on port 179.
        curr_speaker = bgpspeaker.BGPSpeaker(as_number=speaker_as,
                             router_id=self.routerid, bgp_server_port=0,
                             best_path_change_handler=best_path_change_cb,
                             peer_down_handler=bgp_peer_down_cb,
                             peer_up_handler=bgp_peer_up_cb)
        LOG.info(_LI('Added BGP Speaker for local_as=%(as)d with '
                     'router_id= %(rtid)s.'),
                 {'as': speaker_as, 'rtid': self.routerid})

        self.cache.put_bgp_speaker(speaker_as, curr_speaker)

    def delete_bgp_speaker(self, speaker_as):
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)
        # Notify os-ken about BGP Speaker deletion
        curr_speaker.shutdown()
        LOG.info(_LI('Removed BGP Speaker for local_as=%(as)d with '
                     'router_id=%(rtid)s.'),
                 {'as': speaker_as, 'rtid': self.routerid})
        self.cache.remove_bgp_speaker(speaker_as)

    def add_bgp_peer(self, speaker_as, peer_ip, peer_as,
                     auth_type='none', password=None):
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)

        # Validate peer_ip and peer_as.
        utils.validate_as_num('remote_as', peer_as)
        utils.validate_ip_addr(peer_ip)
        utils.validate_auth(auth_type, password)
        if password is not None:
            password = encodeutils.to_utf8(password)

        curr_speaker.neighbor_add(address=peer_ip,
                                  remote_as=peer_as,
                                  enable_ipv4=True,
                                  enable_ipv6=True,
                                  password=password,
                                  connect_mode=CONNECT_MODE_ACTIVE)
        LOG.info(_LI('Added BGP Peer %(peer)s for remote_as=%(as)d to '
                     'BGP Speaker running for local_as=%(local_as)d.'),
                 {'peer': peer_ip, 'as': peer_as, 'local_as': speaker_as})

    def delete_bgp_peer(self, speaker_as, peer_ip):
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)
        # Validate peer_ip. It must be a string.
        utils.validate_ip_addr(peer_ip)

        # Notify os-ken about BGP Peer removal
        curr_speaker.neighbor_del(address=peer_ip)
        LOG.info(_LI('Removed BGP Peer %(peer)s from BGP Speaker '
                     'running for local_as=%(local_as)d.'),
                 {'peer': peer_ip, 'local_as': speaker_as})

    def advertise_route(self, speaker_as, cidr, nexthop):
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)

        # Validate cidr and nexthop. Both must be strings.
        utils.validate_string(cidr)
        utils.validate_string(nexthop)

        # Notify os-ken about route advertisement
        curr_speaker.prefix_add(prefix=cidr, next_hop=nexthop)
        LOG.info(_LI('Route cidr=%(prefix)s, nexthop=%(nexthop)s is '
                     'advertised for BGP Speaker running for '
                     'local_as=%(local_as)d.'),
                 {'prefix': cidr, 'nexthop': nexthop, 'local_as': speaker_as})

    def withdraw_route(self, speaker_as, cidr, nexthop=None):
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)
        # Validate cidr. It must be a string.
        utils.validate_string(cidr)

        # Notify os-ken about route withdrawal
        curr_speaker.prefix_del(prefix=cidr)
        LOG.info(_LI('Route cidr=%(prefix)s is withdrawn from BGP Speaker '
                     'running for local_as=%(local_as)d.'),
                 {'prefix': cidr, 'local_as': speaker_as})

    def get_bgp_speaker_statistics(self, speaker_as):
        LOG.info(_LI('Collecting BGP Speaker statistics for local_as=%d.'),
                 speaker_as)
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)

        # TODO(vikram): Filter and return the necessary information.
        # Will be done as part of new RFE requirement
        # https://bugs.launchpad.net/neutron/+bug/1527993
        return curr_speaker.neighbor_state_get()

    def get_bgp_peer_statistics(self, speaker_as, peer_ip):
        LOG.info(_LI('Collecting BGP Peer statistics for peer_ip=%(peer)s, '
                     'running in speaker_as=%(speaker_as)d '),
                 {'peer': peer_ip, 'speaker_as': speaker_as})
        curr_speaker = self.cache.get_bgp_speaker(speaker_as)
        if not curr_speaker:
            raise bgp_driver_exc.BgpSpeakerNotAdded(local_as=speaker_as,
                                                    rtid=self.routerid)

        # TODO(vikram): Filter and return the necessary information.
        # Will be done as part of new RFE requirement
        # https://bugs.launchpad.net/neutron/+bug/1527993
        return curr_speaker.neighbor_state_get(address=peer_ip)



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import context as nl_context
from neutron_lib.db import api as db_api
from neutron_lib.objects import registry as obj_reg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from sqlalchemy import sql

from neutron.agent.common import utils
from neutron.db.models import agent as agent_model
from neutron.scheduler import base_resource_filter
from neutron.scheduler import base_scheduler

from neutron_dynamic_routing._i18n import _LI, _LW
from neutron_dynamic_routing.api.rpc.callbacks import resources as dr_resources
from neutron_dynamic_routing.db import bgp_db
from neutron_dynamic_routing.db import bgp_dragentscheduler_db as bgp_dras_db
from neutron_dynamic_routing.services.bgp.common import constants as bgp_consts

LOG = logging.getLogger(__name__)
BGP_SPEAKER_PER_DRAGENT = 1


class BgpDrAgentFilter(base_resource_filter.BaseResourceFilter):

    def bind(self, context, agents, bgp_speaker_id, force_scheduling=False):
        """Bind the BgpSpeaker to a BgpDrAgent."""
        bound_agents = agents[:]
        for agent in agents:
            # saving agent_id to use it after rollback to avoid
            # DetachedInstanceError
            agent_id = agent.id
            binding = bgp_dras_db.BgpSpeakerDrAgentBinding()
            binding.agent_id = agent_id
            binding.bgp_speaker_id = bgp_speaker_id
            try:
                with db_api.CONTEXT_WRITER.using(context):
                    context.session.add(binding)
            except db_exc.DBDuplicateEntry:
                # it's totally ok, someone just did our job!
                bound_agents.remove(agent)
                LOG.info(_LI('BgpDrAgent %s already present'), agent_id)
            LOG.debug('BgpSpeaker %(bgp_speaker_id)s is scheduled to be '
                      'hosted by BgpDrAgent %(agent_id)s',
                      {'bgp_speaker_id': bgp_speaker_id,
                       'agent_id': agent_id})
        super(BgpDrAgentFilter, self).bind(context, bound_agents,
                                           bgp_speaker_id, force_scheduling)

    def filter_agents(self, plugin, context, bgp_speaker):
        """Return the agents that can host the BgpSpeaker."""
        agents_dict = self._get_bgp_speaker_hostable_dragents(
            plugin, context, bgp_speaker)
        if not agents_dict['hostable_agents'] or agents_dict['n_agents'] <= 0:
            return {'n_agents': 0,
                    'hostable_agents': [],
                    'hosted_agents': []}
        return agents_dict

    def _get_active_dragents(self, plugin, context):
        """Return a list of active BgpDrAgents."""
        with db_api.CONTEXT_READER.using(context):
            active_dragents = plugin.get_agent_objects(
                context, filters={
                    'agent_type': [bgp_consts.AGENT_TYPE_BGP_ROUTING],
                    'admin_state_up': [True]})
            if not active_dragents:
                return []
        return active_dragents

    def _get_num_dragents_hosting_bgp_speaker(self, bgp_speaker_id,
                                              dragent_bindings):
        return sum(1 if dragent_binding.bgp_speaker_id == bgp_speaker_id else 0
                   for dragent_binding in dragent_bindings)

    def _get_bgp_speaker_hostable_dragents(self, plugin, context, bgp_speaker):
        """Return number of additional BgpDrAgents which will actually host
           the given BgpSpeaker and a list of BgpDrAgents which can host the
           given BgpSpeaker
        """
        # only one BgpSpeaker can be hosted by a BgpDrAgent for now.
        dragents_per_bgp_speaker = BGP_SPEAKER_PER_DRAGENT
        dragent_bindings = plugin.get_dragent_bgp_speaker_bindings(context)
        agents_hosting = [dragent_binding.agent_id
                          for dragent_binding in dragent_bindings]

        num_dragents_hosting_bgp_speaker = (
            self._get_num_dragents_hosting_bgp_speaker(bgp_speaker['id'],
                                                       dragent_bindings))
        n_agents = dragents_per_bgp_speaker - num_dragents_hosting_bgp_speaker
        if n_agents <= 0:
            return {'n_agents': 0,
                    'hostable_agents': [],
                    'hosted_agents': []}

        active_dragents = self._get_active_dragents(plugin, context)
        hostable_dragents = [
            agent for agent in set(active_dragents)
            if agent.id not in agents_hosting and plugin.is_eligible_agent(
                active=True, agent=agent)
        ]
        if not hostable_dragents:
            return {'n_agents': 0,
                    'hostable_agents': [],
                    'hosted_agents': []}

        n_agents = min(len(hostable_dragents), n_agents)
        return {'n_agents': n_agents,
                'hostable_agents': hostable_dragents,
                'hosted_agents': num_dragents_hosting_bgp_speaker}


class BgpDrAgentSchedulerBase(BgpDrAgentFilter):

    periodic_actions = True

    def _register_callbacks(self):
        registry.subscribe(self.schedule_bgp_speaker_callback,
                           dr_resources.BGP_SPEAKER,
                           events.AFTER_CREATE)

    def schedule_all_unscheduled_bgp_speakers(self, context):
        """Call schedule_unscheduled_bgp_speakers for all hosts.
        """

        with db_api.CONTEXT_WRITER.using(context):
            query = context.session.query(agent_model.Agent.host).distinct()
            for agent in query:
                self.schedule_unscheduled_bgp_speakers(context, agent[0])
        return True

    def schedule_unscheduled_bgp_speakers(self, context, host):
        """Schedule unscheduled BgpSpeaker to a BgpDrAgent.
        """

        LOG.debug('Started auto-scheduling on host %s', host)
        with db_api.CONTEXT_WRITER.using(context):
            bgp_dragent = obj_reg.load_class('Agent').get_object(
                context,
                agent_type=bgp_consts.AGENT_TYPE_BGP_ROUTING,
                host=host,
                admin_state_up=True)
            if not bgp_dragent:
                LOG.debug('No enabled BgpDrAgent on host %s', host)
                return False

            if utils.is_agent_down(
                    bgp_dragent.heartbeat_timestamp):
                LOG.warning(_LW('BgpDrAgent %s is down'), bgp_dragent.id)
                return False

            if self._is_bgp_speaker_hosted(context, bgp_dragent['id']):
                # One BgpDrAgent can only host one BGP speaker
                LOG.debug('BgpDrAgent already hosting a speaker on host %s. '
                          'Cannot schedule an another one', host)
                return False

            unscheduled_speakers = self._get_unscheduled_bgp_speakers(context)
            if not unscheduled_speakers:
                LOG.debug('Nothing to auto-schedule on host %s', host)
                return False

            self.bind(context, [bgp_dragent], unscheduled_speakers[0])
        return True

    def _is_bgp_speaker_hosted(self, context, agent_id):
        speaker_binding_model = bgp_dras_db.BgpSpeakerDrAgentBinding

        query = context.session.query(speaker_binding_model)
        query = query.filter(speaker_binding_model.agent_id == agent_id)

        return query.count() > 0

    def _get_unscheduled_bgp_speakers(self, context):
        """BGP speakers that needs to be scheduled.
        """

        no_agent_binding = ~sql.exists().where(
            bgp_db.BgpSpeaker.id ==
            bgp_dras_db.BgpSpeakerDrAgentBinding.bgp_speaker_id)
        query = context.session.query(bgp_db.BgpSpeaker.id).filter(
            no_agent_binding)
        return [bgp_speaker_id_[0] for bgp_speaker_id_ in query]

    def schedule_bgp_speaker_callback(self, resource, event, trigger, payload):
        plugin = payload['plugin']
        if event == events.AFTER_CREATE:
            ctx = nl_context.get_admin_context()
            plugin.schedule_bgp_speaker(ctx, payload['bgp_speaker'])


class ChanceScheduler(base_scheduler.BaseChanceScheduler,
                      BgpDrAgentSchedulerBase):

    def __init__(self):
        super(ChanceScheduler, self).__init__(self)
        self._register_callbacks()


class WeightScheduler(base_scheduler.BaseWeightScheduler,
                      BgpDrAgentSchedulerBase):

    def __init__(self):
        super(WeightScheduler, self).__init__(self)
        self._register_callbacks()


class StaticScheduler(base_scheduler.BaseScheduler,
                      BgpDrAgentFilter):

    periodic_actions = False

    def schedule_all_unscheduled_bgp_speakers(self, context):
        return True

    def schedule_unscheduled_bgp_speakers(self, context, host):
        return True

    def select(self, plugin, context, resource_hostable_agents,
               resource_hosted_agents, num_agents_needed):

        return []



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron.tests import base
from neutron_lib import context

from neutron_dynamic_routing.api.rpc.agentnotifiers import bgp_dr_rpc_agent_api


class TestBgpDrAgentNotifyApi(base.BaseTestCase):

    def setUp(self):
        super(TestBgpDrAgentNotifyApi, self).setUp()
        self.notifier = (
            bgp_dr_rpc_agent_api.BgpDrAgentNotifyApi())

        mock_cast_p = mock.patch.object(self.notifier,
                                        '_notification_host_cast')
        self.mock_cast = mock_cast_p.start()
        mock_call_p = mock.patch.object(self.notifier,
                                        '_notification_host_call')
        self.mock_call = mock_call_p.start()
        self.context = context.get_admin_context()
        self.host = 'host-1'

    def test_agent_updated(self):
        admin_state_up = True
        host = 'my-hostname'
        self.notifier.agent_updated(self.context, admin_state_up, host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)

    def test_notify_dragent_bgp_routes_advertisement(self):
        bgp_speaker_id = 'bgp-speaker-1'
        routes = [{'destination': '1.1.1.1', 'next_hop': '2.2.2.2'}]
        self.notifier.bgp_routes_advertisement(self.context, bgp_speaker_id,
                                               routes, self.host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)

    def test_notify_dragent_bgp_routes_withdrawal(self):
        bgp_speaker_id = 'bgp-speaker-1'
        routes = [{'destination': '1.1.1.1'}]
        self.notifier.bgp_routes_withdrawal(self.context, bgp_speaker_id,
                                            routes, self.host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)

    def test_notify_bgp_peer_disassociated(self):
        bgp_speaker_id = 'bgp-speaker-1'
        bgp_peer_ip = '1.1.1.1'
        self.notifier.bgp_peer_disassociated(self.context, bgp_speaker_id,
                                             bgp_peer_ip, self.host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)

    def test_notify_bgp_peer_associated(self):
        bgp_speaker_id = 'bgp-speaker-1'
        bgp_peer_id = 'bgp-peer-1'
        self.notifier.bgp_peer_associated(self.context, bgp_speaker_id,
                                          bgp_peer_id, self.host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)

    def test_notify_bgp_speaker_created(self):
        bgp_speaker_id = 'bgp-speaker-1'
        self.notifier.bgp_speaker_created(self.context, bgp_speaker_id,
                                          self.host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)

    def test_notify_bgp_speaker_removed(self):
        bgp_speaker_id = 'bgp-speaker-1'
        self.notifier.bgp_speaker_removed(self.context, bgp_speaker_id,
                                          self.host)
        self.assertEqual(1, self.mock_cast.call_count)
        self.assertEqual(0, self.mock_call.call_count)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_policy import policy

from neutron_dynamic_routing.policies import base


rules = [
    policy.DocumentedRuleDefault(
        'add_bgp_speaker_to_dragent',
        base.RULE_ADMIN_ONLY,
        'Add a BGP speaker to a dynamic routing agent',
        [
            {
                'method': 'POST',
                'path': '/agents/{agent_id}/bgp-drinstances',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        'remove_bgp_speaker_from_dragent',
        base.RULE_ADMIN_ONLY,
        'Remove a BGP speaker from a dynamic routing agent',
        [
            {
                'method': 'DELETE',
                'path': '/agents/{agent_id}/bgp-drinstances/{bgp_speaker_id}',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        'list_bgp_speaker_on_dragent',
        base.RULE_ADMIN_ONLY,
        'List BGP speakers hosted by a dynamic routing agent',
        [
            {
                'method': 'GET',
                'path': '/agents/{agent_id}/bgp-drinstances',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        'list_dragent_hosting_bgp_speaker',
        base.RULE_ADMIN_ONLY,
        'List dynamic routing agents hosting a BGP speaker',
        [
            {
                'method': 'GET',
                'path': '/bgp-speakers/{bgp_speaker_id}/bgp-dragents',
            },
        ]
    ),
]


def list_rules():
    return rules



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netaddr

from neutron_lib import constants as lib_consts

from neutron_dynamic_routing.services.bgp.agent.driver import exceptions as bgp_driver_exc  # noqa
from neutron_dynamic_routing.services.bgp.common import constants as bgp_consts  # noqa


# Parameter validation functions provided are provided by the base.
def validate_as_num(param, as_num):
    if not isinstance(as_num, int):
        raise bgp_driver_exc.InvalidParamType(param=param,
                                              param_type='integer')

    if not (bgp_consts.MIN_ASNUM <= as_num <= bgp_consts.MAX_4BYTE_ASNUM):
        # Must be in [AS_NUM_MIN, MAX_4BYTE_ASNUM] range.
        allowed_range = ('[' +
                         str(bgp_consts.MIN_ASNUM) + '-' +
                         str(bgp_consts.MAX_4BYTE_ASNUM) +
                         ']')
        raise bgp_driver_exc.InvalidParamRange(param=param,
                                               range=allowed_range)


def validate_auth(auth_type, password):
    validate_string(password)
    if auth_type in bgp_consts.SUPPORTED_AUTH_TYPES:
        if auth_type != 'none' and password is None:
            raise bgp_driver_exc.PasswordNotSpecified(auth_type=auth_type)
        if auth_type == 'none' and password is not None:
            raise bgp_driver_exc.InvaildAuthType(auth_type=auth_type)
    else:
        raise bgp_driver_exc.InvaildAuthType(auth_type=auth_type)


def validate_ip_addr(ip_addr):
    if netaddr.valid_ipv4(ip_addr):
        return lib_consts.IP_VERSION_4
    elif netaddr.valid_ipv6(ip_addr):
        return lib_consts.IP_VERSION_6
    else:
        raise bgp_driver_exc.InvalidParamType(param=ip_addr,
                                              param_type='ip-address')


def validate_string(param):
    if param is not None:
        if not isinstance(param, str):
            raise bgp_driver_exc.InvalidParamType(param=param,
                                                  param_type='string')


class BgpMultiSpeakerCache(object):
    """Class for saving multiple BGP speakers information.

    Version history:
        1.0 - Initial version for caching multiple BGP speaker information.
    """
    def __init__(self):
        self.cache = {}

    def get_hosted_bgp_speakers_count(self):
        return len(self.cache)

    def put_bgp_speaker(self, local_as, speaker):
        self.cache[local_as] = speaker

    def get_bgp_speaker(self, local_as):
        return self.cache.get(local_as)

    def remove_bgp_speaker(self, local_as):
        self.cache.pop(local_as, None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from neutron_dynamic_routing.policies import bgp_dragent
from neutron_dynamic_routing.policies import bgp_peer
from neutron_dynamic_routing.policies import bgp_speaker


def list_rules():
    return itertools.chain(
        bgp_speaker.list_rules(),
        bgp_peer.list_rules(),
        bgp_dragent.list_rules(),
    )



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib.api.definitions import bgp as bgp_ext
from neutron_lib.plugins import directory
import oslo_messaging


class BgpSpeakerRpcCallback(object):
    """BgpDrAgent RPC callback in plugin implementations.

    This class implements the server side of an RPC interface.
    The client side of this interface can be found in
    neutron_dynamic_routing.services.bgp.agent.bgp_dragent.BgpDrPluginApi.
    For more information about changing RPC interfaces,
    see https://docs.openstack.org/neutron/latest/
    contributor/internals/rpc_api.html.
    """

    # API version history:
    # 1.0 BGPDRPluginApi BASE_RPC_API_VERSION
    target = oslo_messaging.Target(version='1.0')

    @property
    def plugin(self):
        if not hasattr(self, '_plugin'):
            self._plugin = directory.get_plugin(bgp_ext.ALIAS)
        return self._plugin

    def get_bgp_speaker_info(self, context, bgp_speaker_id):
        """Return BGP Speaker details such as peer list and local_as.

        Invoked by the BgpDrAgent to lookup the details of a BGP Speaker.
        """
        return self.plugin.get_bgp_speaker_with_advertised_routes(
                                context, bgp_speaker_id)

    def get_bgp_peer_info(self, context, bgp_peer_id):
        """Return BgpPeer details such as IP, remote_as, and credentials.

        Invoked by the BgpDrAgent to lookup the details of a BGP peer.
        """
        return self.plugin.get_bgp_peer(context, bgp_peer_id,
                                        ['peer_ip', 'remote_as',
                                         'auth_type', 'password'])

    def get_bgp_speakers(self, context, host=None, **kwargs):
        """Returns the list of all BgpSpeakers.

        Typically invoked by the BgpDrAgent as part of its bootstrap process.
        """
        return self.plugin.get_bgp_speakers_for_agent_host(context, host)



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class BgpDriverBase(object, metaclass=abc.ABCMeta):
    """Base class for BGP Speaking drivers.

    Any class which provides BGP functionality should extend this
    defined base class.
    """

    @abc.abstractmethod
    def add_bgp_speaker(self, speaker_as):
        """Add a BGP speaker.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :raises: BgpSpeakerAlreadyScheduled, BgpSpeakerMaxScheduled,
                 InvalidParamType, InvalidParamRange
        """

    @abc.abstractmethod
    def delete_bgp_speaker(self, speaker_as):
        """Deletes BGP speaker.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :raises: BgpSpeakerNotAdded
        """

    @abc.abstractmethod
    def add_bgp_peer(self, speaker_as, peer_ip, peer_as,
                     auth_type='none', password=None):
        """Add a new BGP peer.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :param peer_ip: Specifies the IP address of the peer.
        :type peer_ip: string
        :param peer_as: Specifies Autonomous Number of the peer.
                        Must be an biginteger between MIN_ASNUM and
                        MAX_4BYTE_ASNUM.
        :type peer_as: biginteger
        :param auth_type: Specifies authentication type.
                          By default, authentication will be disabled.
        :type auth_type: value in SUPPORTED_AUTH_TYPES
        :param password: Authentication password.By default, authentication
                         will be disabled.
        :type password: string
        :raises: BgpSpeakerNotAdded, InvalidParamType, InvalidParamRange,
                 InvaildAuthType, PasswordNotSpecified
        """

    @abc.abstractmethod
    def delete_bgp_peer(self, speaker_as, peer_ip):
        """Delete a BGP peer associated with the given peer IP

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :param peer_ip: Specifies the IP address of the peer. Must be the
                        string representation of an IP address.
        :type peer_ip: string
        :raises: BgpSpeakerNotAdded, BgpPeerNotAdded
        """

    @abc.abstractmethod
    def advertise_route(self, speaker_as, cidr, nexthop):
        """Add a new prefix to advertise.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :param cidr: CIDR of the network to advertise. Must be the string
                     representation of an IP network (e.g., 10.1.1.0/24)
        :type cidr: string
        :param nexthop: Specifies the next hop address for the above
                        prefix.
        :type nexthop: string
        :raises: BgpSpeakerNotAdded, InvalidParamType
        """

    @abc.abstractmethod
    def withdraw_route(self, speaker_as, cidr, nexthop=None):
        """Withdraw an advertised prefix.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :param cidr: CIDR of the network to withdraw. Must be the string
                     representation of an IP network (e.g., 10.1.1.0/24)
        :type cidr: string
        :param nexthop: Specifies the next hop address for the above
                        prefix.
        :type nexthop: string
        :raises: BgpSpeakerNotAdded, RouteNotAdvertised, InvalidParamType
        """

    @abc.abstractmethod
    def get_bgp_speaker_statistics(self, speaker_as):
        """Collect BGP Speaker statistics.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :raises: BgpSpeakerNotAdded
        :returns: bgp_speaker_stats: string
        """

    @abc.abstractmethod
    def get_bgp_peer_statistics(self, speaker_as, peer_ip, peer_as):
        """Collect BGP Peer statistics.

        :param speaker_as: Specifies BGP Speaker autonomous system number.
                           Must be an biginteger between MIN_ASNUM and
                           MAX_4BYTE_ASNUM.
        :type speaker_as: biginteger
        :param peer_ip: Specifies the IP address of the peer.
        :type peer_ip: string
        :param peer_as: Specifies the AS number of the peer. Must be an
                        biginteger between MIN_ASNUM and MAX_4BYTE_ASNUM.
        :type peer_as: biginteger                    .
        :raises: BgpSpeakerNotAdded, BgpPeerNotAdded
        :returns: bgp_peer_stats: string
        """



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import rpc as n_rpc
import oslo_messaging

from neutron_dynamic_routing.services.bgp.common import constants as bgp_consts


class BgpDrAgentNotifyApi(object):
    """API for plugin to notify BGP DrAgent.

    This class implements the client side of an rpc interface.  The server side
    is neutron_dynamic_routing.services.bgp.agent.bgp_dragent.BgpDrAgent. For
    more information about rpc interfaces, please see
    https://docs.openstack.org/neutron/latest/contributor/internals/rpc_api.html.
    """

    def __init__(self, topic=bgp_consts.BGP_DRAGENT):
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)
        self.topic = topic

    def agent_updated(self, context, admin_state_up, host):
        """Tell BgpDrAgent that agent was updated.

        This effectively tells the bgp_dragent to resync.
        """
        self._notification_host_cast(context, 'agent_updated',
                {'admin_state_up': admin_state_up}, host)

    def bgp_routes_advertisement(self, context, bgp_speaker_id,
                                 routes, host):
        """Tell BgpDrAgent to begin advertising the given route.

        Invoked on FIP association, adding router port to a tenant network,
        and new DVR port-host bindings, and subnet creation(?).
        """
        self._notification_host_cast(context, 'bgp_routes_advertisement_end',
                {'advertise_routes': {'speaker_id': bgp_speaker_id,
                                      'routes': routes}}, host)

    def bgp_routes_withdrawal(self, context, bgp_speaker_id,
                              routes, host):
        """Tell BgpDrAgent to stop advertising the given route.

        Invoked on FIP disassociation, removal of a router port on a
        network, and removal of DVR port-host binding, and subnet delete(?).
        """
        self._notification_host_cast(context, 'bgp_routes_withdrawal_end',
                {'withdraw_routes': {'speaker_id': bgp_speaker_id,
                                     'routes': routes}}, host)

    def bgp_peer_disassociated(self, context, bgp_speaker_id,
                               bgp_peer_ip, host):
        """Tell BgpDrAgent about a BGP Peer disassociation.

        This effectively tells the BgpDrAgent to stop a peering session.
        """
        self._notification_host_cast(context, 'bgp_peer_disassociation_end',
                {'bgp_peer': {'speaker_id': bgp_speaker_id,
                              'peer_ip': bgp_peer_ip}}, host)

    def bgp_peer_associated(self, context, bgp_speaker_id,
                            bgp_peer_id, host):
        """Tell BgpDrAgent about a new BGP Peer association.

        This effectively tells the bgp_dragent to open a peering session.
        """
        self._notification_host_cast(context, 'bgp_peer_association_end',
                {'bgp_peer': {'speaker_id': bgp_speaker_id,
                              'peer_id': bgp_peer_id}}, host)

    def bgp_speaker_created(self, context, bgp_speaker_id, host):
        """Tell BgpDrAgent about the creation of a BGP Speaker.

        Because a BGP Speaker can be created with BgpPeer binding in place,
        we need to inform the BgpDrAgent of a new BGP Speaker in case a
        peering session needs to opened immediately.
        """
        self._notification_host_cast(context, 'bgp_speaker_create_end',
                {'bgp_speaker': {'id': bgp_speaker_id}}, host)

    def bgp_speaker_removed(self, context, bgp_speaker_id, host):
        """Tell BgpDrAgent about the removal of a BGP Speaker.

        Because a BGP Speaker can be removed with BGP Peer binding in
        place, we need to inform the BgpDrAgent of the removal of a
        BGP Speaker in case peering sessions need to be stopped.
        """
        self._notification_host_cast(context, 'bgp_speaker_remove_end',
                {'bgp_speaker': {'id': bgp_speaker_id}}, host)

    def _notification_host_cast(self, context, method, payload, host):
        """Send payload to BgpDrAgent in the cast mode"""
        cctxt = self.client.prepare(topic=self.topic, server=host)
        cctxt.cast(context, method, payload=payload)

    def _notification_host_call(self, context, method, payload, host):
        """Send payload to BgpDrAgent in the call mode"""
        cctxt = self.client.prepare(topic=self.topic, server=host)
        cctxt.call(context, method, payload=payload)



# Copyright (C) 2017 VA Linux Systems Japan K.K.
# Copyright (C) 2017 Fumihiko Kakuma <kakuma at valinux co jp>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron.tests import base
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const

from neutron_dynamic_routing.api.rpc.callbacks import resources as dr_resources
from neutron_dynamic_routing.db import bgp_db
from neutron_dynamic_routing.services.bgp import bgp_plugin


class TestBgpPlugin(base.BaseTestCase):

    def setUp(self):
        super(TestBgpPlugin, self).setUp()
        bgp_notify_p = mock.patch('neutron_dynamic_routing.api.rpc.'
                                  'agentnotifiers.bgp_dr_rpc_agent_api.'
                                  'BgpDrAgentNotifyApi')
        bgp_notify_p.start()
        rpc_conn_p = mock.patch('neutron_lib.rpc.Connection')
        rpc_conn_p.start()
        admin_ctx_p = mock.patch('neutron_lib.context.get_admin_context')
        self.admin_ctx_m = admin_ctx_p.start()
        self.fake_admin_ctx = mock.Mock()
        self.admin_ctx_m.return_value = self.fake_admin_ctx
        self.plugin = bgp_plugin.BgpPlugin()

    def _create_test_payload(self, context='test_ctx'):
        bgp_speaker = {'id': '11111111-2222-3333-4444-555555555555'}
        payload = {'plugin': self.plugin, 'context': context,
                   'bgp_speaker': bgp_speaker}
        return payload

    def test__register_callbacks(self):
        with mock.patch.object(registry, 'subscribe') as subscribe:
            plugin = bgp_plugin.BgpPlugin()
            expected_calls = [
                mock.call(plugin.bgp_drscheduler.schedule_bgp_speaker_callback,
                          dr_resources.BGP_SPEAKER, events.AFTER_CREATE),
                mock.call(plugin.floatingip_update_callback,
                          resources.FLOATING_IP, events.AFTER_CREATE),
                mock.call(plugin.floatingip_update_callback,
                          resources.FLOATING_IP, events.AFTER_UPDATE),
                mock.call(plugin.router_interface_callback,
                          resources.ROUTER_INTERFACE, events.AFTER_CREATE),
                mock.call(plugin.router_interface_callback,
                          resources.ROUTER_INTERFACE, events.BEFORE_CREATE),
                mock.call(plugin.router_interface_callback,
                          resources.ROUTER_INTERFACE, events.AFTER_DELETE),
                mock.call(plugin.router_gateway_callback,
                          resources.ROUTER_GATEWAY, events.AFTER_CREATE),
                mock.call(plugin.router_gateway_callback,
                          resources.ROUTER_GATEWAY, events.AFTER_DELETE),
                mock.call(plugin.port_callback,
                          resources.PORT, events.AFTER_UPDATE),
            ]
            self.assertEqual(subscribe.call_args_list, expected_calls)

    def test_create_bgp_speaker(self):
        test_context = 'create_bgp_context'
        test_bgp_speaker = {'id': None}
        payload = self._create_test_payload(context=test_context)
        with mock.patch.object(bgp_db.BgpDbMixin,
                               'create_bgp_speaker') as create_bgp_sp:
            with mock.patch.object(registry, 'notify') as notify:
                create_bgp_sp.return_value = payload['bgp_speaker']
                self.assertEqual(self.plugin.create_bgp_speaker(
                    test_context, test_bgp_speaker),
                    payload['bgp_speaker'])
                create_bgp_sp.assert_called_once_with(test_context,
                                                      test_bgp_speaker)
                notify.assert_called_once_with(dr_resources.BGP_SPEAKER,
                                               events.AFTER_CREATE,
                                               self.plugin,
                                               payload=payload)

    def test_floatingip_update_callback(self):
        new_fip = {'floating_ip_address': netaddr.IPAddress('10.10.10.10'),
                   'router_id': '', 'floating_network_id': 'a-b-c-d-e'}
        old_fip = new_fip.copy()
        old_fip.update(router_id='old-router-id')

        test_context = 'test_context'

        get_bpg_speakers_name = '_bgp_speakers_for_gw_network_by_family'
        with mock.patch.object(self.plugin, get_bpg_speakers_name) as get_bgp:
            with mock.patch.object(self.plugin,
                                   'stop_route_advertisements') as stop_ad:
                with mock.patch.object(self.plugin, '_bgp_rpc') as bgp_rpc:
                    bgp_speaker = mock.Mock()
                    bgp_speaker.id = '11111111-2222-3333-4444-555555555555'
                    get_bgp.return_value = [bgp_speaker]

                    self.plugin.floatingip_update_callback(
                        test_context, events.AFTER_UPDATE, None,
                        payload=events.DBEventPayload(
                            test_context, states=(old_fip, new_fip)))

                    get_bgp.assert_called_once_with(self.fake_admin_ctx,
                                                    'a-b-c-d-e',
                                                    n_const.IP_VERSION_4)

                    old_host_route = [{'destination': '10.10.10.10/32',
                        'next_hop': None}]
                    stop_ad.assert_called_once_with(self.fake_admin_ctx,
                                                    bgp_rpc,
                                                    bgp_speaker.id,
                                                    old_host_route)
